### Based on Sun, J., Sinha, S., Wang, S., and Maity, T. (SIM 2011)
### Bias corrected inference for the conditional logistic regression(CLR)
### Extended to P-dimensional covariate vector

#========= Expression function in CLR =======================
ef4 <- function(beta, vi, X){
   X <- as.matrix(X)
   t <- exp(X%*%beta)
   s <- sum(vi * t)
   return(s)
} #Expression


#========= Information matrix ===============================
# M: number of controls
infor4 <- function(beta, z, x, X,ncase, M){
   s1 <- 0 
   for(i in 1:ncase){ 
       index <- ((i-1)*M + i):(i*M + i)
       zi <- z[index]
       xi <- x[index]
	 Xi <- X[index,]
       t <- ef4(beta, rep(1,M+1), Xi ) 
       z1 <- ef4(beta, zi,   Xi)
       x1 <- ef4(beta, xi,   Xi)

       xz <-  ef4(beta, xi*zi, Xi ) 
       s1 <- s1 + xz/t - z1*x1/t^2
    } 
    return(s1)
} #return an element in information matrix

Infor4matrix <- function(beta, X,ncase, M){
   P <- ncol(X)
   A <- matrix(0, P, P)
   for (i in 1: P) {
      for (j in i:P) {
         z = X[,i]
         x = X[,j]
         a<- infor4(beta, z, x, X,ncase, M)
         A[i,j] <- A[j,i] <- a
       }
    }
   return(A)
} #return a P*P information matrix


#============Dibeta=====( Dibeta1, Dibeta2, ...)================
# contribute to first derivative of information matrix
Dibeta4 <- function(beta, z, x,w, X, ncase, M){
    s2 <- 0
    for(i in 1:ncase){
       index <- ((i-1)*M + i):(i*M + i)
       zi <- z[index]
       xi <- x[index]
       wi <- w[index]
	 Xi <- X[index,]
      
       t <- ef4(beta, rep(1,M+1), Xi )  
       z1 <- ef4(beta, zi, Xi)
       x1 <- ef4(beta, xi, Xi)
       w1 <- ef4(beta, wi, Xi)
       zx <- ef4(beta, zi*xi, Xi)
       zw <- ef4(beta, zi*wi, Xi)
       xw <- ef4(beta, xi*wi, Xi)
       zxw <- ef4(beta, zi*xi*wi, Xi)

	 #Dbeta1_beta2_beta3
       s2 <- s2+zxw/t-zx*w1/t^2-(zw*x1+z1*xw)/t^2+2*z1*x1*w1/t^3   
     }
   return(s2)
} #return an element for first derivative of information matrix

Dibeta4matrix <- function(beta, i, X, ncase, M){
    P <- ncol(X)
    z <- X[,i]
    B <- matrix(0, P, P)
    for (j in 1:P) {
        for (k in j:P) {
            x <- X[,j]
            w <- X[,k]
            B[j,k] <- B[k,j] <- Dibeta4(beta, z, x, w, X, ncase, M)
        }
    }
    return(B)
} #return a P*P matrix for first derivative of information matrix



#============DDibeta=====( DDibeta1, DDibeta2, ...)================
# contribute to second derivative of information matrix
DDibeta4 <- function(beta, z, x, w, v, X, ncase, M){
     s3 <- 0
     for(i in 1:ncase){
       index <- ((i-1)*M + i):(i*M + i)
       zi <- z[index]
       xi <- x[index]
       wi <- w[index]
       vi <- v[index]
	 Xi <- X[index,]
       
       t <- ef4(beta, rep(1,M+1), Xi )  
       z1 <- ef4(beta, zi, Xi)
       x1 <- ef4(beta, xi, Xi)
       w1 <- ef4(beta, wi, Xi)
       v1 <- ef4(beta, vi, Xi)
       zx <- ef4(beta, zi*xi, Xi)
       zw <- ef4(beta, zi*wi, Xi)
       zv <- ef4(beta, zi*vi, Xi)
       xw <- ef4(beta, wi*xi, Xi)
       xv <- ef4(beta, vi*xi, Xi)
       wv <- ef4(beta, wi*vi, Xi)

       zxw <- ef4(beta, zi*xi*wi, Xi)
       zxv <- ef4(beta, zi*xi*vi, Xi)
       zwv <- ef4(beta, zi*vi*wi, Xi)
       xwv <- ef4(beta, xi*vi*wi, Xi)
      zxwv <- ef4(beta, zi*xi*wi*vi, Xi)

	#Dbeta1_beta2_beta3 _beta4 
	s3 <- s3+ zxwv/t-zxw*v1/t^2-(zxv*w1+zx*wv)/t^2+2*zx*w1*v1/t^3-(zwv*x1+zw*xv+zv*xw+z1*xwv)/t^2+2*(zw*x1+xw*z1)*v1/t^3+2*(zv*x1*w1+z1*xv*w1+z1*x1*wv)/t^3-6*(z1*x1*w1*v1)/t^4
	 }  
     return(s3)
} #return an element for second derivative of information matrix

DDibeta4matrix <- function(beta, i, ii, X,ncase, M){
   P <- ncol(X)
   z <- X[,i]
   x <- X[,ii]

   D <- matrix(0, P, P)
      for (j in 1:P) {
         for (k in j:P) {
            w <- X[,j]
            v <- X[,k]
            D[j,k] <- D[k,j] <- DDibeta4(beta, z, x, w, v, X,ncase, M)
         }
      }
   return(D)
}#return a P*P matrix for second derivative of information matrix


#==================== S(beta):Score function ================================
Scor4 <- function(beta, y, z, X, ncase, M){
      s <- 0
      for(i in 1:ncase){
         index <- ((i-1)*M + i):(i*M + i)
         zi <- z[index]
    	   Xi <- X[index,]
         yi <- y[index]
         t <- ef4(beta, rep(1,M+1), Xi ) 
         z1 <- ef4( beta, zi , Xi )
         s <- s + sum(yi*zi) - z1/t
     	}
     return(s)
} #return an element in Score function

Scor4matrix <- function(beta, y, X, ncase, M){
	P <- ncol(X)
	S <- matrix(0, nrow=P, ncol=1)
	for (i in 1: P) {
		z <- X[,i]
		S[i,] <- Scor4(beta,y, z, X,ncase, M)
	}
return(S)
} #return a P*1 matrix of Score function


#============= S.star(beta): modified Score function =============================================
Scor.star4 <- function(beta, y, X, ncase, M){
	P <- ncol(X)
      itheta <- Infor4matrix(beta, X, ncase, M)
      S <- Scor4matrix(beta, y, X, ncase, M)
      t0 <- solve(itheta)
	
	S_star <- matrix(0,nrow=P, ncol=1)
	for ( i in 1:P) {
      	H1 <- Dibeta4matrix (beta, i,  X, ncase, M)  
      	t1 <- t0%*%H1
      	S_star[i,] <- S[i] +  sum(diag(t1))/2 
	}
      return(S_star)
} #return a P*1 matrix of modified Score function


#=============Si(beta):ith component of socre funtion S(beta)============================================
Si4 <- function(beta, yi, zi, X, M){
      t  <- ef4(beta, rep(1,M+1), X)
      z1 <- ef4(beta, zi, X)
      v1 <- sum(yi*zi) - z1/t
      return(v1)
} #element of function "Si4matrix"

Si4matrix <- function(beta, y, X, M){
	P <- ncol(X)
	Si <- vector()
	for (i in 1: P) {
		z = X[,i]
		Si[i]<- Si4(beta, y, z, X, M)
	}
	return(Si)
}#return a vector 


#========= SST_star(beta):sum of Si(beta) * Si(beta)^T  ===============================
SST.star4 <- function(beta, y, X, ncase, M){
      itheta <- Infor4matrix(beta, X, ncase, M)
      t0 <- solve(itheta)

	tra <- vector()
	P <- ncol(X)
	for ( p in 1:P) {
      	H1 <- Dibeta4matrix (beta, p, X, ncase, M)  
      	t1 <- t0%*%H1
      	tra[p] <- sum(diag(t1))/(2*ncase)
	 }

      v <- 0
      for(i in 1:ncase){
         index <- ((i-1)*M + i):(i*M + i)
         yi <- y[index]
         Xi <- X[index,]
         v0 <- Si4matrix(beta, yi, Xi, M) 
         v <- v + v0%*%t(tra) + tra%*%t(v0) 
      } #v is a P*P matrix

	return( v + itheta + ncase*tra%*%t(tra) )
} #return a P*P matrix 

#========= Trace matrix ===========================================
Mtracr4 <- function(beta, X, ncase, M){
	P <- ncol(X)
	I <- Infor4matrix(beta, X, ncase, M)
	ivI <- solve(I)

	AA <- matrix(0, nrow=P, ncol=P)
	for ( i in 1:P) {
      	b1 <- Dibeta4matrix (beta, i,  X, ncase, M)  
      	bb1 <- DDibeta4matrix (beta, i, i, X, ncase, M)  
      	a11 <- sum(diag(-1*ivI%*%b1%*%ivI%*%b1 + ivI %*% bb1))
      	AA[i,i] <- a11
	 }

	for ( i in 1:(P-1)) {
		for (j in (i+1):P) {
      		b1 <- Dibeta4matrix (beta, i,  X, ncase, M)  
      		b2 <- Dibeta4matrix (beta, j,  X, ncase, M)  
      		b12 <- DDibeta4matrix (beta, i, j, X, ncase, M)  
      		a12 <- sum( diag(-1*ivI%*%b2%*%ivI%*%b1 + ivI %*% b12))
     			AA[i,j] <- AA[j,i] <- a12
		}
	}
     return(AA)
} #return a P*P matrix of trace matrix of information matrix


#===================== Newton Method to solve S.star(beta)=0 ==============================
Newton4 <- function(eps, beta, y, X, ncase, M){
      P <- ncol(X)
      beta1 <- matrix(0, nrow=P, ncol=1)
      beta <- matrix(beta, nrow=P, ncol=1)
      flag <- 0
      dist <- 1
      ite <- 0
      ite_max <- 25
      if(max(abs(beta))>=8 ) flag <- 99 else{
          S.star <- Scor.star4(beta, y, X, ncase, M)
          H <- Infor4matrix(beta, X, ncase, M)
          MM <- Mtracr4(beta, X, ncase, M)
          I.star <- H - MM/2
          eg <- eigen(I.star)$values 
          if(min(abs(eg)) <= 0.0001) flag <- 77
       }

      while( sqrt(dist) > eps & flag==0){ 
              beta1 <- beta + solve(I.star)%*% S.star
              dist <- sum((beta1-beta)^2)
              beta <- beta1
              if(max(abs(beta))>=8) flag <- 99 else {
                 H <- Infor4matrix(beta, X, ncase, M)
                 MM <- Mtracr4(beta, X, ncase, M)
                 I.star <- H - MM/2
                 eg <- eigen(I.star)$values 
                 if(min(abs(eg)) <= 0.0001) flag <- 77  else {
                      S.star <- Scor.star4(beta, y, X, ncase, M)
                      ite <- ite + 1
                   if(ite > ite_max & sqrt(dist)>eps) flag <-888
                 }
               }
      }   #while  
      if(flag==99 || flag==77) return(c(9999,9999)) else if(flag==888) {return(c(888,888))}
      else return(beta)
}# end of function "Newton4" for P-dim covariate

#=================================================================================================
# The follow code is copied from original code in http://www.stat.tamu.edu/~sinha/research.html
#=============== Bias correction for univariate covariate ========================================
ef <- function(beta, vi, xi){
   t <- exp(xi*beta)
   s <- sum(vi * t)
   return(s)
}
#================( Score funtion S(beta), Information matrix i(beta), Di_Dbeta, DDi_DDbeta )==========

Dibeta <- function(beta,y, x, ncase, M){
     s1 <- 0  # S(beta)
     s2 <- 0  # I(beta) 
     s3 <- 0  # first derivative of i(beta)
     s4 <- 0  # second derivative of i(beta) 
       
    for(i in 1:ncase){
       index <- ((i-1)*M + i):(i*M + i)
       xi <- x[index]
 	 yi <- y[index]
       t  <- ef(beta, rep(1,M+1), xi) 
       x1 <- ef(beta, xi,   xi)
       x2 <- ef(beta, xi^2, xi)  
       x3 <- ef(beta, xi^3, xi)
       x4 <- ef(beta, xi^4, xi)      
  
       s1 <- s1 + sum(yi*xi) - x1/t   # score  function 
       s2 <- s2 + x2/t - (x1/t)^2  # information matrix
       s3 <- s3 + x3/t - x2*x1/t^2 - 2*x1*(x2*t-x1^2)/t^3  #-DlogL_Dbeta^3 = Di_beta
       m1 <- (x3*t - x1*x2)*t^2 - 2*(x2*t-x1^2)*t*x1
       s4 <- s4 + x4/t - (x2/t)^2 - 2*x1*( x3/t^2 - x2*x1/t^3) - 2*( (x2*t-x1^2)^2/t^4 + x1*m1/t^5 )  #-DlogL_Dbeta^4=Di_beta^2
    }
       v <- c(s1, s2, s3, s4)
       return(v)
}  

#============= S.star: Corrected score function ====================
Scor.star <- function(beta, y, x, ncase, M){
      H <- Dibeta(beta, y, x, ncase, M)
      v <- H[1] + 1/2 * H[3]/H[2]
      return(v)
}

#=============Si(theta): ith component of socre funtion S(beta) ===========
Si <- function(beta, yi, xi, M){
      t  <- ef(beta, rep(1,M+1), xi) 
      x1 <- ef(beta, xi , xi) 
      v <- sum(yi*xi) - x1/t    
      return(v)
}

#========= SST_star(theta): sum of Si.star * Si.star^T ==================
SST.star <- function(beta, y, x, ncase, M){
      H <- Dibeta(beta, y, x, ncase, M)
      m <- 1/(2*ncase) * H[3]/H[2]
      s <- 0
      for(i in 1:ncase){
         index <- ((i-1)*M + i):(i*M + i)
         yi <- y[index]
         xi <- x[index]
         tt <- Si(beta, yi, xi, M)  
         s <- s + tt %*% t(m) + m %*% t(tt)
      }
      return(s + H[2] + ncase * m%*%t(m))
}

#===================== Newton Method to solve S.star(beta)=0 ==========================
Newton <- function(eps, beta, y, x, ncase, M){
      flag <- 0
      dist <- 1
      ite <- 0
      ite_max <- 100
      s.star <- Scor.star(beta, y, x, ncase, M)
      if(abs(beta)>=20) return(9999)  else if(abs(s.star) <= 1.0e-8) return(beta) else{ 
          H <- Dibeta(beta, y, x, ncase, M)
          MM <- H[4]/H[2] - (H[3]/H[2])^2 
          I.star <- H[2] - MM/2
          if(abs(I.star) <= 0.0001) flag=77 }
      
      while( dist> eps & abs(s.star)>1.0e-8 & flag==0 ){ 
              beta1 <- beta + s.star/I.star
              dist <- abs(beta1-beta)
              beta <- beta1
              if(abs(beta)>=20) flag=99 else{ 
                  H <- Dibeta(beta, y, x, ncase, M)
                  MM <- H[4]/H[2] - (H[3]/H[2])^2 
                  I.star <- H[2] - MM/2
                  if(abs(I.star) <= 0.0001) flag <- 77  else {
                     s.star <- Scor.star(beta, y, x, ncase, M)
                     ite <- ite + 1
                     if(ite > ite_max ) flag <-888 
                  }
              }
      }   #while  
      if(flag==99|| flag==77) return(9999) else if(flag==888) {return(888)} else return(beta)
} #end of Newton for univaraite



#===================================================================
# Main function:
MDS2 <- function(y, covariate, ncase, M, eps, beta0){
    if(ncol(covariate)>1){
     	 estimate <- Newton4(eps, beta0, y, covariate, ncase, M)
    	 uu <- Infor4matrix(estimate, covariate, ncase, M) 
    	 Mtr <- Mtracr4(estimate, covariate, ncase, M)
    	 T <- -uu + Mtr/2
    	 DSS <- solve(T)
    	 SST0 <- SST.star4(estimate, y, covariate, ncase, M)
   	 mm <- diag(DSS %*% SST0 %*% DSS)   # Expression (4) in Sun et al. (2011)
    	 std <- sqrt(mm)
       return(data.frame(estimate, std)) 
	} 

   if(ncol(covariate)==1){ 
     	 estimate <- Newton(eps, beta0, y, covariate, ncase, M)
       H <- Dibeta(estimate,y, covariate, ncase, M) 
       uu <- H[2]
       Mtr <- H[4]/H[2]-(H[3]/H[2])^2
       T <-  -uu + Mtr/2
       DSS <- 1/T
       SST0 <- SST.star(estimate, y, covariate, ncase, M)
       mm <- DSS*SST0*DSS 
       std <- sqrt(mm)
       return(data.frame(estimate, std)) 
      }
} # end of programming
